﻿unit UDLight;

interface

procedure Register;

implementation

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Classes,
  System.Generics.Collections, System.Generics.Defaults, Vcl.Graphics, ToolsAPI,
  DebugAPI, System.Rtti, Vcl.Forms, System.TypInfo, Vcl.ExtCtrls, UDLight.DDetours;

type
  TIDENotifier = class(TNotifierObject, IOTAIDENotifier)
  private
    FEditorNotifiers: TList<IOTAEditorNotifier>;
  public
    constructor Create;
    destructor Destroy; override;

    { IOTAIDENotifier }
    procedure FileNotification(NotifyCode: TOTAFileNotification;
      const FileName: string; var Cancel: Boolean);
    procedure BeforeCompile(const Project: IOTAProject; var Cancel: Boolean); overload;
    procedure AfterCompile(Succeeded: Boolean); overload;
  end;

  TEditorNotifier = class(TNotifierObject, IOTANotifier, IOTAEditorNotifier)
  private
    FSourceEditor: IOTASourceEditor;
    FEditViewNotifiers: TList<INTAEditViewNotifier>;
    FNotifierIndex: Integer;
    procedure RemoveNotifier;
  public
    constructor Create(ASourceEditor: IOTASourceEditor);
    destructor Destroy; override;

    { IOTANotifier }
    procedure Destroyed;
    procedure IOTANotifier.Destroyed = Destroyed;
    { IOTAEditorNotifier }
    procedure ViewNotification(const View: IOTAEditView; Operation: TOperation);
    procedure ViewActivated(const View: IOTAEditView);
  end;

  TEditViewNotifier = class(TNotifierObject, IOTANotifier, INTAEditViewNotifier)
  private
    FEditView: IOTAEditView;
    FNotifierIndex: Integer;
    procedure RemoveNotifier;
  public
    constructor Create(AEditView: IOTAEditView);
    destructor Destroy; override;

    { IOTANotifier }
    procedure Destroyed;
    procedure IOTANotifier.Destroyed = Destroyed;
    { INTAEditViewNotifier }
    procedure EditorIdle(const View: IOTAEditView);
    procedure BeginPaint(const View: IOTAEditView; var FullRepaint: Boolean);
    procedure PaintLine(const View: IOTAEditView; LineNumber: Integer;
      const LineText: PAnsiChar; const TextWidth: Word; const LineAttributes: TOTAAttributeArray;
      const Canvas: TCanvas; const TextRect: TRect; const LineRect: TRect; const CellSize: TSize);
    procedure EndPaint(const View: IOTAEditView);
  end;

  TDebuggerNotifier = class(TNotifierObject, IOTADebuggerNotifier, IOTADebuggerNotifier90)
  public
    { IOTADebuggerNotifier }
    procedure ProcessCreated(const Process: IOTAProcess);
    procedure ProcessDestroyed(const Process: IOTAProcess);
    procedure BreakpointAdded(const Breakpoint: IOTABreakpoint);
    procedure BreakpointDeleted(const Breakpoint: IOTABreakpoint);
    { IOTADebuggerNotifier90 }
    procedure BreakpointChanged(const Breakpoint: IOTABreakpoint);
    procedure CurrentProcessChanged(const Process: IOTAProcess);
    procedure ProcessStateChanged(const Process: IOTAProcess);
    function BeforeProgramLaunch(const Project: IOTAProject): Boolean;
    procedure ProcessMemoryChanged;
  end;

var
  FDLightEnabled: Boolean;
  FDebugger: IOTADebugger;
  FIDENotifierIndex: Integer = -1;
  FDebuggerNotifierIndex: Integer = -1;
  FLocalVariables: TDictionary<string,string>;
  FWatchExpressions: TList<TPair<string,string>>;
  FCurrentBuffer: IOTAEditBuffer;
  FRepaintAll: Boolean;
  FHideLocalVarIfLineContainsExpr: Boolean = True;
  FRepaintTimer: TTimer;
  FTWatchWindow_GetWatchExpression: function(Self: TObject; Index: Integer): string;
  FTWatchWindow_GetWatchItemCount: function(Self: TObject): Integer;
  FTrampolineWatchWindowEvaluateWatches: procedure(Self: TObject; var Message: TMessage);
  FOriginalWindowProc: TWndMethod;
  FWatchWindow: TForm;
{$IFDEF DEBUG}
  FEditorNotifierCount: Integer;
  FEditViewNotifierCount: Integer;
{$ENDIF}

{ TIDENotifier }

constructor TIDENotifier.Create;
var
  moduleServices: IOTAModuleServices;
  i, j: Integer;
  module: IOTAModule;
  editor: IOTASourceEditor;
begin
  inherited;
  FEditorNotifiers := TList<IOTAEditorNotifier>.Create;

  moduleServices := BorlandIDEServices as IOTAModuleServices;
  for i := 0 to moduleServices.ModuleCount-1 do
  begin
    module := moduleServices.Modules[i];

    for j := 0 to module.ModuleFileCount-1 do
      if Supports(module.ModuleFileEditors[j], IOTASourceEditor, editor) then
        FEditorNotifiers.Add(TEditorNotifier.Create(editor));
  end;
end;

destructor TIDENotifier.Destroy;
var
  i: Integer;
begin
  for i := 0 to FEditorNotifiers.Count-1 do
    FEditorNotifiers[i].Destroyed;
  FEditorNotifiers.Free;
  inherited;
end;

procedure TIDENotifier.FileNotification(NotifyCode: TOTAFileNotification;
  const FileName: string; var Cancel: Boolean);
var
  module: IOTAModule;
  i: Integer;
  editor: IOTASourceEditor;
begin
  if NotifyCode = ofnFileOpened then
  begin
    module := (BorlandIDEServices as IOTAModuleServices).FindModule(FileName);
    if not Assigned(Module) then Exit;
    for i := 0 to module.ModuleFileCount-1 do
      if Supports(module.ModuleFileEditors[i], IOTASourceEditor, editor) then
        FEditorNotifiers.Add(TEditorNotifier.Create(editor));
  end;
end;

procedure TIDENotifier.BeforeCompile(const Project: IOTAProject; var Cancel: Boolean);
begin
end;

procedure TIDENotifier.AfterCompile(Succeeded: Boolean);
begin
end;

{ TEditorNotifier }

constructor TEditorNotifier.Create(ASourceEditor: IOTASourceEditor);
var
  i: Integer;
begin
  inherited Create;
  FEditViewNotifiers := TList<INTAEditViewNotifier>.Create;
  FSourceEditor := ASourceEditor;

  FNotifierIndex := FSourceEditor.AddNotifier(Self);
{$IFDEF DEBUG}
  Inc(FEditorNotifierCount);
{$ENDIF}
  for i := 0 to FSourceEditor.EditViewCount-1 do
  begin
    FEditViewNotifiers.Add(TEditViewNotifier.Create(FSourceEditor.EditViews[i]));
  end;

end;

destructor TEditorNotifier.Destroy;
var
  i: Integer;
begin
  for i := 0 to FEditViewNotifiers.Count-1 do
    FEditViewNotifiers[i].Destroyed;
  FEditViewNotifiers.Free;
  RemoveNotifier;
  inherited;
end;

procedure TEditorNotifier.Destroyed;
begin
  RemoveNotifier;
end;

procedure TEditorNotifier.ViewNotification(const View: IOTAEditView; Operation: TOperation);
begin
  if Operation = opInsert then
    FEditViewNotifiers.Add(TEditViewNotifier.Create(View));
end;

procedure TEditorNotifier.ViewActivated(const View: IOTAEditView);
begin
end;

procedure TEditorNotifier.RemoveNotifier;
begin
  if Assigned(FSourceEditor) and (FNotifierIndex >= 0) then
  begin
    FSourceEditor.RemoveNotifier(FNotifierIndex);
    FNotifierIndex := -1;
    FSourceEditor := nil;
{$IFDEF DEBUG}
    Dec(FEditorNotifierCount);
{$ENDIF}
  end;
end;

{ TEditViewNotifier }

constructor TEditViewNotifier.Create(AEditView: IOTAEditView);
begin
  inherited Create;
  FEditView := AEditView;
  FNotifierIndex := FEditView.AddNotifier(Self);
{$IFDEF DEBUG}
  Inc(FEditViewNotifierCount);
{$ENDIF}
end;

destructor TEditViewNotifier.Destroy;
begin
  RemoveNotifier;
  inherited;
end;

procedure TEditViewNotifier.Destroyed;
begin
  RemoveNotifier;
end;

procedure TEditViewNotifier.EditorIdle(const View: IOTAEditView);
begin
end;

procedure TEditViewNotifier.BeginPaint(const View: IOTAEditView; var FullRepaint: Boolean);
begin
  if FRepaintAll then
  begin
    FullRepaint := True;
    FRepaintAll := False;
  end;
end;

function GetIdentifiers(const LineText: PAnsiChar; const TextWidth: Word;
  const LineAttributes: TOTAAttributeArray): TArray<string>;
var
  isIdentifier: Boolean;
  i: Integer;
  p, p1: PAnsiChar;
  identA: AnsiString;
  identU: string;
  idents: TStringList;
begin
  idents := TStringlist.Create;
  try
    idents.CaseSensitive := False;

    isIdentifier := False;
    p := LineText;
    p1 := LineText;
    for i := Low(LineAttributes) to High(LineAttributes) do
    begin
      case LineAttributes[i] of
        atIdentifier:
        begin
          if not isIdentifier then
          begin
            isIdentifier := True;
            p1 := p;
          end
        end;
      else
        if isIdentifier then
        begin
          isIdentifier := False;

          SetString(identA, p1, p - p1);
          identU := string(identA);
          if idents.IndexOf(identU) < 0 then
            idents.Add(identU);
        end;
      end;
      Inc(p);
    end;
    if isIdentifier then
    begin
      SetString(identA, p1, p - p1);
      identU := string(identA);
      if idents.IndexOf(identU) < 0 then
        idents.Add(identU);
    end;

    Result := idents.ToStringArray;
  finally
    idents.Free;
  end;
end;

procedure TEditViewNotifier.PaintLine(const View: IOTAEditView; LineNumber: Integer;
  const LineText: PAnsiChar; const TextWidth: Word; const LineAttributes: TOTAAttributeArray;
  const Canvas: TCanvas; const TextRect: TRect; const LineRect: TRect; const CellSize: TSize);
var
  i, p: Integer;
  idents: TArray<string>;
  localValue, lineValue: string;
  list: TList<TPair<Integer,string>>;
  lineTextUtf8, exprUtf8: UTF8String;
begin
//  if not FEnableDLight then
//  begin
//    Canvas.Brush.Style := bsClear;
//    Canvas.TextOut(TextRect.Right, TextRect.Top, '💩');
//    Canvas.Brush.Style := bsSolid;
//  end;

  if not FDLightEnabled then
    Exit;
  if (FCurrentBuffer = nil) or (View.Buffer <> FCurrentBuffer) then
    Exit;
  if (FLocalVariables.Count = 0) and (FWatchExpressions.Count = 0) then
    Exit;

  // Watch expressions
  lineTextUtf8 := UTF8String(LowerCase(string(UTF8String(LineText))));
  list := TList<TPair<Integer,string>>.Create;
  try
    for i := 0 to FWatchExpressions.Count-1 do
    begin
      exprUtf8 := UTF8String(FWatchExpressions[i].Key.ToLower);
      p := Pos(exprUtf8, lineTextUtf8);
      if p = 0 then Continue;
      if LineAttributes[p-1] <> atIdentifier then Continue;
      list.Add(TPair<Integer,string>.Create(p, FWatchExpressions[i].Key + '=' + FWatchExpressions[i].Value));
    end;
    list.Sort(TComparer<TPair<Integer,string>>.Construct(
      function(const Left, Right: TPair<Integer,string>): Integer
      begin
        Result := Left.Key - Right.Key;
      end));
    for i := 0 to list.Count-1 do
    begin
      if lineValue = '' then
        lineValue := list[i].Value
      else
        lineValue := lineValue + ' ' + list[i].Value;
    end;
  finally
    list.Free;
  end;

  // Local variables
  lineTextUtf8 := UTF8String(LowerCase(string(UTF8String(LineText))));
  if not FHideLocalVarIfLineContainsExpr or (lineValue = '') then
  begin
    idents := GetIdentifiers(LineText, TextWidth, LineAttributes);
    for i := Low(idents) to High(idents) do
    begin
      if FLocalVariables.TryGetValue(LowerCase(idents[i]), localValue) then
      begin
        if lineValue = '' then
          lineValue := idents[i] + '=' + localValue
        else
          lineValue := lineValue + ' ' + idents[i] + '=' + localValue;
      end;
    end;
  end;

  if lineValue = '' then Exit;

  Canvas.Brush.Color := clTeal;
  Canvas.Font.Color := clAqua;
  Canvas.TextOut(TextRect.Right+10, TextRect.Top, lineValue);
end;

procedure TEditViewNotifier.EndPaint(const View: IOTAEditView);
begin
end;

procedure TEditViewNotifier.RemoveNotifier;
begin
  if Assigned(FEditView) and (FNotifierIndex >= 0) then
  begin
    FEditView.RemoveNotifier(FNotifierIndex);
    FNotifierIndex := -1;
    FEditView := nil;
{$IFDEF DEBUG}
    Dec(FEditViewNotifierCount);
{$ENDIF}
  end;
end;

{ TDebuggerNotifier }

procedure TDebuggerNotifier.ProcessCreated(const Process: IOTAProcess);
begin
end;

procedure TDebuggerNotifier.ProcessDestroyed(const Process: IOTAProcess);
begin
  FCurrentBuffer := nil;
  FLocalVariables.Clear;
  FWatchExpressions.Clear;
end;

procedure TDebuggerNotifier.BreakpointAdded(const Breakpoint: IOTABreakpoint);
begin
end;

procedure TDebuggerNotifier.BreakpointDeleted(const Breakpoint: IOTABreakpoint);
begin
end;

procedure TDebuggerNotifier.BreakpointChanged(const Breakpoint: IOTABreakpoint);
begin
end;

procedure TDebuggerNotifier.CurrentProcessChanged(const Process: IOTAProcess);
begin
end;

procedure GetLocalVariables;
var
  debuggerLocalVariables: IOTADebuggerLocalVariables;
  expressionInspector: IOTAExpressionInspector;
  i, memberCount: Integer;
  localName, localValue: string;
begin
  FLocalVariables.Clear;
  if not FDLightEnabled then
    Exit;
  if not Supports(FDebugger, IOTADebuggerLocalVariables, debuggerLocalVariables) then
    Exit;
  if debuggerLocalVariables.InspectLocalVariables(expressionInspector) <> erOK then
    Exit;

  memberCount := expressionInspector.MemberCount[eimtData];
  for i := 0 to memberCount-1 do
  begin
    localName := LowerCase(expressionInspector.MemberName[eimtData, i]);
    localValue := expressionInspector.MemberValue[eimtData, i];
    FLocalVariables.Add(localName, localValue);
  end;
end;

procedure GetWatchExpressions;
var
  debuggerInspector: IOTADebuggerInspector;
  expressionInspector: IOTAExpressionInspector;
  i, itemCount: Integer;
  watchExpr, watchValue: string;
begin
  FWatchExpressions.Clear;
  if not FDLightEnabled then
    Exit;
  if not Supports(FDebugger, IOTADebuggerInspector, debuggerInspector) then
    Exit;
  if not Assigned(FTWatchWindow_GetWatchExpression) or not Assigned(FTWatchWindow_GetWatchItemCount) then
    Exit;
  if not Assigned(FWatchWindow) then
    Exit;

  itemCount := FTWatchWindow_GetWatchItemCount(FWatchWindow);
  for i := 0 to itemCount-1 do
  begin
    watchExpr := FTWatchWindow_GetWatchExpression(FWatchWindow, i);
    if debuggerInspector.CreateInspector(watchExpr, expressionInspector) = erOK then
    begin
      watchValue := expressionInspector.Value;
      FWatchExpressions.Add(TPair<string,string>.Create(watchExpr, watchValue));
    end;
  end;
end;

procedure TDebuggerNotifier.ProcessStateChanged(const Process: IOTAProcess);
var
  view: IOTAEditView;
begin
  FDLightEnabled := False;

  view := (BorlandIDEServices as IOTAEditorServices).TopView;
  if view = nil then
  begin
    FCurrentBuffer := nil;
    FLocalVariables.Clear;
    FWatchExpressions.Clear;
    Exit;
  end;

  if Process.ProcessState <> psStopped then
  begin
    FRepaintTimer.Enabled := True;
    FLocalVariables.Clear;
    FWatchExpressions.Clear;
    Exit;
  end;

  FDLightEnabled := True;
  FCurrentBuffer := view.Buffer;
  FRepaintTimer.Enabled := False;

  GetLocalVariables;
  GetWatchExpressions;
end;

function TDebuggerNotifier.BeforeProgramLaunch(const Project: IOTAProject): Boolean;
begin
  Result := True;
end;

procedure TDebuggerNotifier.ProcessMemoryChanged;
begin
end;

procedure Register;
begin
  FDebugger := DebugAPI.DebuggerManagerServices.CurrentDebugger;
  if FDebugger = nil then Exit;

  FDebuggerNotifierIndex := FDebugger.AddNotifier(TDebuggerNotifier.Create);
  FIDENotifierIndex := (BorlandIDEServices as IOTAServices).AddNotifier(TIDENotifier.Create);

{$IFDEF DEBUG}
  OutputDebugString('DLight installed');
{$ENDIF}
end;

procedure Unregister;
begin
  if FDebugger = nil then Exit;

  if FDebuggerNotifierIndex >= 0 then
    FDebugger.RemoveNotifier(FDebuggerNotifierIndex);
  if FIDENotifierIndex >= 0 then
    (BorlandIDEServices as IOTAServices).RemoveNotifier(FIDENotifierIndex);

{$IFDEF DEBUG}
  OutputDebugString('DLight uninstalled');
{$ENDIF}
end;

procedure DoRepaint;
var
  i: Integer;
begin
  FRepaintAll := True;
  if FCurrentBuffer <> nil then
  begin
    for i := 0 to FCurrentBuffer.EditViewCount-1 do
      FCurrentBuffer.EditViews[i].Paint;
  end;
end;

procedure TimerRepaint(Self: TTimer; Sender: TObject);
begin
  Self.Enabled := False;
  DoRepaint;
  FCurrentBuffer := nil;
end;

function CreateMethod(Self: TObject; Proc: Pointer): TMethod;
begin
  TMethod(Result).Code := Proc;
  TMethod(Result).Data := Self;
end;

procedure WatchWindowProc(Self: TForm; var Message: TMessage);
begin
  if (Message.Msg = $50D) and FDLightEnabled then
  begin
    GetWatchExpressions;
    DoRepaint;
  end;
  FOriginalWindowProc(Message);
end;

procedure MyWatchWindowEvaluateWatches(Self: TObject; var Message: TMessage);
begin
  if FDLightEnabled then
  begin
    GetWatchExpressions;
    DoRepaint;
  end;
  FTrampolineWatchWindowEvaluateWatches(Self, Message);
end;

procedure GetWatchWindowInfo;
var
  ctx: TRttiContext;
  typ: TRttiType;
  propWatchExpr: TRttiIndexedProperty;
  propWatchCount: TRttiProperty;
begin
  FWatchWindow := TForm(Application.FindComponent('WatchWindow'));
  if FWatchWindow = nil then Exit;

  //typ := ctx.FindType('WatchWin.TWatchWindow');
  typ := ctx.GetType(FWatchWindow.ClassType);
  if typ = nil then Exit;
  propWatchExpr := typ.GetIndexedProperty('WatchExpression');
  if propWatchExpr = nil then Exit;
  propWatchCount := typ.GetProperty('WatchItemCount');
  if propWatchCount = nil then Exit;

  FTWatchWindow_GetWatchExpression := propWatchExpr.ReadMethod.CodeAddress;
  FTWatchWindow_GetWatchItemCount := TRttiInstanceProperty(propWatchCount).PropInfo^.GetProc;

  FOriginalWindowProc := FWatchWindow.WindowProc;
  FWatchWindow.WindowProc := TWndMethod(CreateMethod(FWatchWindow, @WatchWindowProc));

  @FTrampolineWatchWindowEvaluateWatches := InterceptCreate('coreide240.bpl', '@Watchwin@TWatchWindow@EvaluateWatches$qqrr24Winapi@Messages@TMessage', @MyWatchWindowEvaluateWatches);
end;

initialization
  FLocalVariables := TDictionary<string,string>.Create;
  FWatchExpressions := TList<TPair<string,string>>.Create;
  FRepaintTimer := TTimer.Create(nil);
  FRepaintTimer.Enabled := False;
  FRepaintTimer.Interval := 100;
  FRepaintTimer.OnTimer := TNotifyEvent(CreateMethod(FRepaintTimer, @TimerRepaint));
  GetWatchWindowInfo;
finalization
  FRepaintTimer.Free;
  Unregister;
  FLocalVariables.Free;
  FWatchExpressions.Free;
  if Assigned(FWatchWindow) and Assigned(FOriginalWindowProc) then
    FWatchWindow.WindowProc := FOriginalWindowProc;
  if Assigned(FTrampolineWatchWindowEvaluateWatches) then
    InterceptRemove(@FTrampolineWatchWindowEvaluateWatches);

{$IFDEF DEBUG}
  OutputDebugString(PChar(Format('EditorNotifierCount: %d', [FEditorNotifierCount])));
  OutputDebugString(PChar(Format('EditViewNotifierCount: %d', [FEditViewNotifierCount])));
{$ENDIF}
end.
